
/********************************************************************************/
/*                                                                              */
/*  Practica de Arquitectures Orientades a les Aplicacions                      */
/*                                                                              */
/*  Versio SISTOLICA del programa de comparacio de les paraules d'un text       */
/*  amb un mot clau per mitja de la Distancia de Levenshtein.                   */
/*  Implementacio amb C-stolic de l'algorisme de la distancia de Levenshtein.   */
/*  Autors: Francisco Jose Aguilar Celdran                                      */
/*          Sergio Blanco Cuaresma                                              */
/*                                                                              */
/********************************************************************************/

//#include <stdlib.h>
//#include <stdio.h>

#define TAM_MAX_PARAULA 10
#define NUM_CELLES (TAM_MAX_PARAULA*2)-1 

void distancia_levenshtein (char cadena_a[TAM_MAX_PARAULA], char cadena_b[TAM_MAX_PARAULA], int retorn[1])
{
   int i;
   
   systolic int esq;
   systolic int esq_nou;   // Rebem per l'esquerra
   systolic int dret;
   systolic int dret_nou;  // Rebem per la dreta
   systolic int anterior;
   systolic int inici;  // Fins que no sigui 0 no s'inicia les execucions utils
   systolic int execucions_utils;  // Nombre d'execucions utils: es fan calculs valids
   systolic int torn;   // Comportament boolea, les celles fan execucions utils alternes;
   systolic int identificador;   // Cella numero X
   //systolic int cost_substitucio[TAM_MAX_PARAULA];
   systolic int cost_substitucio;
   systolic int cost_substitucioA;
   systolic int cost_substitucioB;
   systolic char paraula1[TAM_MAX_PARAULA];
   systolic char paraula2[TAM_MAX_PARAULA];
   systolic int minim;  // Per calcular el valor minim
   systolic int valor;
   
   // Inicialitzem els torns alternatius i l'identificador de cella   
   for (i = 2; i <= NUM_CELLES+1; i++) {
      torn => torn : i%2;
      identificador => identificador : i-2;
   }
      
   // Inicialitzem el valor dret: 0,0,0,0,0,0,0,0,0,0,1,2,3,4,5,6,7,8,9,10
   for (i = TAM_MAX_PARAULA; i > 0; i--) {
      dret => dret : i;
      dret_nou => dret_nou : i;
   }
   
   for (i = TAM_MAX_PARAULA-1; i > 0; i--) {
      dret => dret : 0;
      dret_nou => dret_nou : 0;
   }
   
   // Inicialitzem el valor esquerra: 10,9,8,7,6,5,4,3,2,1,x,x,x,x,x,x,x,x,x,x
   for (i = 1; i <= TAM_MAX_PARAULA-1; i++) {   // [OPCIONAL?]
      esq => esq : 0;
      esq_nou => esq_nou : 0;
   }
   
   for (i = 1; i <= TAM_MAX_PARAULA; i++) {
      esq => esq : i;
      esq_nou => esq_nou : i;
   }
   
   // Inicialitzem el valor d'inici que indica quan ha de comenar a fer calculs
   // la cella (quan sigui = 0): 9,8,7,6,5,4,3,2,1,0,1,2,3,4,5,6,7,8,9
   for (i = TAM_MAX_PARAULA-1; i >= 0; i--) {
      inici => inici : i;
   }
   
   for (i = 1; i <= TAM_MAX_PARAULA-1; i++) {
      inici => inici : i;
   }
   
   // Inicialitzaci del cost de substitucio
   for (i=0; i<TAM_MAX_PARAULA; i++) {
      paraula1[i] =| cadena_a[i];
      paraula2[i] =| cadena_b[i];
   }
   
   // Primera execuci
   execucions_utils =| 0;
   
   minim =| 0;
   valor =| 0;
   anterior =| 0;
   
   // Calcul de la Distancia de Levenshtein
   for(i = 0; i < NUM_CELLES; i++) {
      
      // Si es la primera execuci s'ha d'actualitzar noms un costat segons
      // on estigui situada la cella
      dret = ((inici == 0) && (identificador < TAM_MAX_PARAULA)) ? dret_nou : dret;
      esq = ((inici == 0) && (identificador > TAM_MAX_PARAULA)) ? esq_nou : esq;
      
      // Noms actualitzar dret/esq si ja hem fet la primera execuci
      dret = (inici < 0) ? dret_nou : dret;
      esq = (inici < 0) ? esq_nou : esq;

      // S'ha de calcular per separat per no machacar dades.
      cost_substitucioA = (((inici <= 0) && (torn == 1)) && (identificador < TAM_MAX_PARAULA) && (paraula1[(TAM_MAX_PARAULA-identificador)+execucions_utils-1] == paraula2[execucions_utils])) ? 0 : 1;
      cost_substitucioB = (((inici <= 0) && (torn == 1)) && (identificador >= TAM_MAX_PARAULA) && (paraula1[execucions_utils] == paraula2[(identificador%TAM_MAX_PARAULA)+execucions_utils])) ? 0 : 1;
      cost_substitucio = (((inici <= 0) && (torn == 1)) && (identificador < TAM_MAX_PARAULA)) ? cost_substitucioA : cost_substitucioB;
      
      // Si ja hem iniciat les execucions i es el nostre torn, trobar el mnim
      //minim = ((inici <= 0) && (torn == 1)) ? anterior + cost_substitucio : minim ;
      minim = ((inici <= 0) && (torn == 1)) ? anterior + cost_substitucio : minim ;
      //minim = ((inici <= 0) && (torn == 1)) ? anterior + 100 : minim ;
      minim = (((inici <= 0) && (torn == 1)) && (dret + 1 < minim)) ? dret + 1 : minim ;
      minim = (((inici <= 0) && (torn == 1)) && (esq + 1 < minim)) ? esq + 1 : minim ;
      
      valor = ((inici <= 0) && (torn == 1)) ? minim : valor;
      anterior = ((inici <= 0) && (torn == 1)) ? valor : anterior;
      
      // Nombre de vegades que ha executat alguna cosa util la cella
      execucions_utils = ((inici <= 0) && (torn == 1)) ? execucions_utils+1 : execucions_utils;
      
      dret_nou =< valor;
      esq_nou => valor;
      
      torn = (torn == 1) ? 0 : 1;
      inici--;
   }
   
   // Recuperem el resultat que es troba a la cella del mig
   for (i = 0; i < TAM_MAX_PARAULA; i++) {
      valor : retorn[0] => valor;
   }
}
