
/********************************************************************************/
/*                                                                              */
/*  Practica de Arquitectures Orientades a les Aplicacions                      */
/*                                                                              */
/*  Versio NO SISTOLICA del programa de comparacio de les paraules d'un text    */
/*  amb un mot clau per mitja de la Distancia de Levenshtein.                   */
/*  Entrades del Programa: - Fitxer de text en format ascii.                    */
/*                         - Paraula clau a cercar.                             */
/*                         - Nombre de paraules amb minima distancia a trobar.  */
/*  Sortides del Programa: - Fitxer de text amb les paraules trobades i les     */
/*                             seves respectives distancies amb el mot clau.    */
/*  Autors: Francisco Jose Aguilar Celdran                                      */
/*          Sergio Blanco Cuaresma                                              */
/*                                                                              */
/********************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#define TAM_MAX_PARAULA 40

struct Info_Paraula
{
   char *nom;
   int distancia;
};
typedef struct Info_Paraula Tipus_Info_Paraula;


/* ------------------------------------------------------------------------------------------- */
/*  DISTANCIA_LEVENSHTEIN: obte la distancia entre dues paraules segons l'esmentat algorisme.  */
/* ------------------------------------------------------------------------------------------- */

int distancia_levenshtein (char *cadena_a, char *cadena_b)
{
   int longitud_a;
   int longitud_b;
   int **matriu;
   int opcio[3];
   int minim;
   int cost;
   int i, j, k;
   int resultat;
   
   longitud_a = strlen(cadena_a);
   longitud_b = strlen(cadena_b);
   if (longitud_a == 0) {
      return longitud_b;
   }
   else if (longitud_b == 0) {
      return longitud_a;
   }
   else {
      // Creacio dinamica de la matriu
      matriu = (int **) malloc((longitud_b + 1) * sizeof(int *));
      for (i = 0; i <= longitud_b; i++) {
         matriu[i] = (int *) malloc((longitud_a + 1) * sizeof(int));
      }
      // Inicialitzacio de la primera fila i columna de la matriu
      for (j = 0; j <= longitud_a; j++) {
         matriu[0][j] = j;
      }
      for (i = 0; i <= longitud_b; i++) {
         matriu[i][0] = i;
      }
      // Calcul de la Distancia de Levenshtein
      for(i = 1; i <= longitud_b; i++) {
         for(j = 1; j <= longitud_a; j++) { 
            if (cadena_a[i - 1] == cadena_b[j - 1]) {
               cost = 0;
            }
            else {
               cost = 1;
            }
            // S'assigna a la cella [i][j] de la matriu el valor menor de les tres opcions
            opcio[0] = matriu[i-1][j] + 1;
            opcio[1] = matriu[i][j-1] + 1;
            opcio[2] = matriu[i-1][j-1] + cost;
            minim = 32500;
            for (k = 0; k < 3; k++) {
               if (opcio[k] < minim) {
                  minim = opcio[k];
               }
            }
            matriu[i][j] = minim;
         }
      }
      resultat = matriu[longitud_b][longitud_a];
      free(matriu);
      return resultat;
   }
}

/* ----------------------------------------------------------------------------------------- */
/*  DISTANCIA_MAXIMA_ACTUAL : obte la posicio de la paraula menys semblant amb el mot clau,  */
/*                            de la taula de paraules semblants.                             */
/* ----------------------------------------------------------------------------------------- */

int distancia_maxima_actual (Tipus_Info_Paraula *semblants, int num_paraules)
{
   int pos_maxima = 0;
   int dist_maxima = 0;
   int i;
   
   for (i = 0; i < num_paraules; i++) {
      if (semblants[i].distancia > dist_maxima) {
         dist_maxima = semblants[i].distancia;
         pos_maxima = i;
      }
   }
   return pos_maxima;
}

/* ----------------------------------------------------------------------------------------- */
/*  ORDENAR_TAULA_PARAULES : ordena per criteri de minima distancia, de menor a major, les   */
/*                           paraules de la taula de paraules semblants.                     */
/* ----------------------------------------------------------------------------------------- */

void ordenar_taula_paraules (Tipus_Info_Paraula *semblants, int num_paraules, int num_actual)
{
   int i, j;
   int minim, min_index;
   char *nom_aux;
   int dist_aux;
   
   if (num_actual < num_paraules) {
      num_paraules = num_actual;
   }
   nom_aux = (char *)malloc(TAM_MAX_PARAULA * sizeof(char));
   for (i = 0; i < num_paraules - 1; i++) {
      minim = semblants[i + 1].distancia;
      min_index = i + 1;
      for (j = i + 2; j < num_paraules; j++) {
         if (semblants[j].distancia < minim) {
             minim = semblants[j].distancia;
             min_index = j;
         }
      }
      if (semblants[i].distancia > minim) {
         strcpy(nom_aux,semblants[min_index]. nom);
         dist_aux = semblants[min_index].distancia;
         semblants[min_index].distancia = semblants[i].distancia;
         strcpy(semblants[min_index].nom, semblants[i].nom);
         semblants[i].distancia = dist_aux;
         strcpy(semblants[i].nom, nom_aux);
      }
   }
   free(nom_aux);
}

/* ---------------------------------------------------------------------------------------------- */
/*  IMPRIMIR_SINTAXIS: Mostra per la sortida estandard la sintaxis dels parametres del programa.  */
/* ---------------------------------------------------------------------------------------------- */

void imprimir_sintaxis (FILE *desti, int mode)
{
   if (mode < 3) {
      fprintf(desti,"\n----------------------------------------------------------------------");
      fprintf(desti,"\nComparacio de cadenes usant la distancia de Levenshtein (Practica AOA)");
      fprintf(desti,"\n----------------------------------------------------------------------");
   }
   if (mode == 1) {
      fprintf(desti,"\nLa sintaxis del programa es: (versio no sistolica)\n");
      fprintf(desti,"\n no_sistolic [fitxer_cerca] [paraula] [num_paraules]\n");
      fprintf(desti,"\n    [fitxer_cerca] -> Nom del fitxer de paraules amb format ascii.");
      fprintf(desti,"\n    [paraula]      -> Paraula a cercar.");
      fprintf(desti,"\n    [num_paraules] -> Numero de paraules a mostrar\n\n");
   }
   else if (mode == 2) {
      fprintf(desti,"\nCalculant les distancies minimes de les paraules del fitxer amb el mot clau...\n");
   }
   else if (mode == 3) {
      fprintf(desti,"\n Fi de l'execucio, resultats emmagatzemats en el fitxer 'mots_semblants.txt'\n\n");      
   }
}

/* ---------------------------------------------------------------------------------------- */
/*  ESCRIURE_RESULTATS: Copia els resultats ordenats de les paraules semblants trobades al  */
/*                      fitxer de sortida amb el nom especificat.                           */
/* ---------------------------------------------------------------------------------------- */

void escriure_resultats (char *fitxer, Tipus_Info_Paraula *semblants, int num, int actual,char *mot)
{
   FILE *sortida;
   int i;
   
   sortida = fopen(fitxer,"w");
   imprimir_sintaxis(sortida,0);
   fprintf(sortida,"\n\nResultats de les %d paraules mes semblants al mot clau '%s':\n\n",num,mot);
   fprintf(sortida,"(Columna esquerra: Nom paraula) (Columna dreta: Distancia amb el mot clau)\n\n");
   for (i = 0; i < actual; i++) {
      fprintf(sortida,"%40s  -->  %d\n",semblants[i].nom, semblants[i].distancia);
   }
   if (actual < num) {
      for (i = actual; i < num; i++) {
         fprintf(sortida,"%40s  -->\n"," ");
      }
   }
   fclose(sortida);
}

/* -------------------------------------------------------------------------------------------- */
/*  PARAULA_REPETIDA: Comprova si la paraula actual ja esta en la taula de paraules semblants.  */
/* -------------------------------------------------------------------------------------------- */

int paraula_repetida (char *paraula, Tipus_Info_Paraula *semblants, int num)
{
   int repetida = 0;
   int i;
   
   for (i = 0; i < num; i++) {
      if (strcmp(semblants[i].nom,paraula) == 0) {
         repetida = 1;
         break;
      }
   }
   return repetida;
}

/* --------------------------------------------------------------------------------------------- */
/*  PROGRAMA PRINCIPAL: Implementa el proces de lectura de totes les paraules del fitxer rebut,  */
/*                      aplicant-hi la distancia de levenshtein, obtenint com a resultat les     */
/*                      'num_paraules' mes semblants al mot clau, amb les seves distancies.      */
/* --------------------------------------------------------------------------------------------- */

int main (int argc, char **argv)
{
   int distancia;
   FILE *fitxer;
   char *mot_clau;
   char *mot_actual;
   Tipus_Info_Paraula *semblants;
   int num_paraules;
   int digits_ok;
   int max_distancia;
   int pos_max_distancia;
   int num_actual;
   int i;
   
   if (argc != 4) {
      imprimir_sintaxis(stdout,1);
      return -1;
   }
   fitxer = fopen(argv[1],"r");
   if (fitxer == NULL) {
      imprimir_sintaxis(stdout,1);
      return -1;
   }
   mot_clau = (char *)malloc((strlen(argv[2]) + 1) * sizeof(char));
   strcpy(mot_clau,argv[2]);
   digits_ok = 1;
   for (i = 0; i < strlen(argv[3]); i++) {
      if (!isdigit(argv[3][i])) {
         digits_ok = 0;
         break;
      }
   }
   if (!digits_ok) {
      imprimir_sintaxis(stdout,1);
      return -1;
   }
   num_paraules = atoi(argv[3]);
   semblants = (Tipus_Info_Paraula *)malloc(num_paraules * sizeof(Tipus_Info_Paraula));
   mot_actual = (char *)malloc(TAM_MAX_PARAULA * sizeof(char));
   imprimir_sintaxis(stdout,2);
   num_actual = 0;
   max_distancia = 0;
   fscanf(fitxer,"%s",&mot_actual[0]);
   
   // Execucio de la distancia de Levenshtein per a cada paraula del fitxer i pel mot clau,
   // classificant les 'num_paraules' amb menor distancia, posteriorment guardades a fitxer.
   while (!feof(fitxer)) {
      distancia = distancia_levenshtein(mot_clau,mot_actual);
      if (num_actual < num_paraules) {
         if (!paraula_repetida(mot_actual,&semblants[0],num_actual)) {
            semblants[num_actual].nom = (char *)malloc(TAM_MAX_PARAULA * sizeof(char));
            strcpy(semblants[num_actual].nom,mot_actual);
            semblants[num_actual].distancia = distancia;
            if (distancia > max_distancia) {
               max_distancia = distancia;
               pos_max_distancia = num_actual;
            }
            num_actual++;
         }
      }
      else {
         if (distancia < max_distancia) {
            if (!paraula_repetida(mot_actual,&semblants[0],num_paraules)) {
               strcpy(semblants[pos_max_distancia].nom,mot_actual);
               semblants[pos_max_distancia].distancia = distancia;
               pos_max_distancia = distancia_maxima_actual(&semblants[0],num_paraules);
               max_distancia = semblants[pos_max_distancia].distancia;
            }
         }
      }
      fscanf(fitxer,"%s",&mot_actual[0]);
   }
   fclose(fitxer);
   ordenar_taula_paraules(&semblants[0],num_paraules,num_actual);
   escriure_resultats("mots_semblants.txt",&semblants[0],num_paraules,num_actual,&mot_clau[0]);
   imprimir_sintaxis(stdout,3);
   free(semblants);
   free(mot_clau);
   free(mot_actual);
   return 1;
}


