
/********************************************************************************/
/*                                                                              */
/*  Practica de Arquitectures Orientades a les Aplicacions                      */
/*                                                                              */
/*  Versio SISTOLICA del programa de comparacio de les paraules d'un text       */
/*  amb un mot clau per mitja de la Distancia de Levenshtein.                   */
/*  Entrades del Programa: - Fitxer de text en format ascii.                    */
/*                         - Paraula clau a cercar.                             */
/*                         - Nombre de paraules amb minima distancia a trobar.  */
/*  Sortides del Programa: - Fitxer de text amb les paraules trobades i les     */
/*                             seves respectives distancies amb el mot clau.    */
/*  ANTENCI: Versi de paraules de mida variable                               */
/*  Autors: Francisco Jose Aguilar Celdran                                      */
/*          Sergio Blanco Cuaresma                                              */
/*                                                                              */
/********************************************************************************/

#include <stdio.h>
#include <stdlib.h>

#define TAM_MAX_PARAULA 10


/* ----------------------------------------------------------------------------------------- */
/*  DISTANCIA_MAXIMA_ACTUAL : obte la posicio de la paraula menys semblant amb el mot clau,  */
/*                            de la taula de paraules semblants.                             */
/* ----------------------------------------------------------------------------------------- */

int distancia_maxima_actual (int *dist_semblants, int num_paraules)
{
   int pos_maxima = 0;
   int dist_maxima = 0;
   int i;
   
   for (i = 0; i < num_paraules; i++) {
      if (dist_semblants[i] > dist_maxima) {
         dist_maxima = dist_semblants[i];
         pos_maxima = i;
      }
   }
   return pos_maxima;
}

/* ----------------------------------------------------------------------------------------- */
/*  ORDENAR_TAULA_PARAULES : ordena per criteri de minima distancia, de menor a major, les   */
/*                           paraules de la taula de paraules semblants.                     */
/* ----------------------------------------------------------------------------------------- */

void ordenar_taula_paraules (char **noms_semblants, int *dist_semblants, int num_paraules, int num_actual)
{
   int i, j;
   int minim, min_index;
   char *nom_aux;
   int dist_aux;
   
   if (num_actual < num_paraules) {
      num_paraules = num_actual;
   }
   nom_aux = (char *)malloc(TAM_MAX_PARAULA * sizeof(char));
   for (i = 0; i < num_paraules - 1; i++) {
      minim = dist_semblants[i + 1];
      min_index = i + 1;
      for (j = i + 2; j < num_paraules; j++) {
         if (dist_semblants[j] < minim) {
             minim = dist_semblants[j];
             min_index = j;
         }
      }
      if (dist_semblants[i] > minim) {
         strcpy(nom_aux,noms_semblants[min_index]);
         dist_aux = dist_semblants[min_index];
         dist_semblants[min_index] = dist_semblants[i];
         strcpy(noms_semblants[min_index], noms_semblants[i]);
         dist_semblants[i] = dist_aux;
         strcpy(noms_semblants[i], nom_aux);
      }
   }
   free(nom_aux);
}

/* ---------------------------------------------------------------------------------------------- */
/*  IMPRIMIR_SINTAXIS: Mostra per la sortida estandard la sintaxis dels parametres del programa.  */
/* ---------------------------------------------------------------------------------------------- */

void imprimir_sintaxis (FILE *desti, int mode)
{
   if (mode < 3) {
      fprintf(desti,"\n----------------------------------------------------------------------");
      fprintf(desti,"\nComparacio de cadenes usant la distancia de Levenshtein (Practica AOA)");
      fprintf(desti,"\n----------------------------------------------------------------------");
   }
   if (mode == 1) {
      fprintf(desti,"\nLa sintaxis del programa es: (versio no sistolica)\n");
      fprintf(desti,"\n no_sistolic [fitxer_cerca] [paraula] [num_paraules]\n");
      fprintf(desti,"\n    [fitxer_cerca] -> Nom del fitxer de paraules amb format ascii.");
      fprintf(desti,"\n    [paraula]      -> Paraula a cercar.");
      fprintf(desti,"\n    [num_paraules] -> Numero de paraules a mostrar\n\n");
   }
   else if (mode == 2) {
      fprintf(desti,"\nCalculant les distancies minimes de les paraules del fitxer amb el mot clau...\n");
   }
   else if (mode == 3) {
      fprintf(desti,"\n Fi de l'execucio, resultats emmagatzemats en el fitxer 'mots_semblants.txt'\n\n");      
   }
}

/* ---------------------------------------------------------------------------------------- */
/*  ESCRIURE_RESULTATS: Copia els resultats ordenats de les paraules semblants trobades al  */
/*                      fitxer de sortida amb el nom especificat.                           */
/* ---------------------------------------------------------------------------------------- */

void escriure_resultats (char *fitxer, char **noms_semblants, int *dist_semblants, int num, int actual, char *mot)
{
   FILE *sortida;
   int i;
   
   sortida = fopen(fitxer,"w");
   imprimir_sintaxis(sortida,0);
   fprintf(sortida,"\n\nResultats de les %d paraules mes semblants al mot clau '%s':\n\n",num,mot);
   fprintf(sortida,"(Columna esquerra: Nom paraula) (Columna dreta: Distancia amb el mot clau)\n\n");
   for (i = 0; i < actual; i++) {
      fprintf(sortida,"%40s  -->  %d\n", noms_semblants[i], dist_semblants[i]);
   }
   if (actual < num) {
      for (i = actual; i < num; i++) {
         fprintf(sortida,"%40s  -->\n"," ");
      }
   }
   fclose(sortida);
}

/* -------------------------------------------------------------------------------------------- */
/*  PARAULA_REPETIDA: Comprova si la paraula actual ja esta en la taula de paraules semblants.  */
/* -------------------------------------------------------------------------------------------- */

int paraula_repetida (char *paraula, char **noms_semblants, int num)
{
   int repetida = 0;
   int i;
   
   for (i = 0; i < num; i++) {
      if (strcmp(noms_semblants[i],paraula) == 0) {
         repetida = 1;
         break;
      }
   }
   return repetida;
}

/* --------------------------------------------------------------------------------------------- */
/*  PROGRAMA PRINCIPAL: Implementa el proces de lectura de totes les paraules del fitxer rebut,  */
/*                      aplicant-hi la distancia de levenshtein, obtenint com a resultat les     */
/*                      'num_paraules' mes semblants al mot clau, amb les seves distancies.      */
/* --------------------------------------------------------------------------------------------- */

int main (int argc, char **argv)
{
   int distancia;
   FILE *fitxer;
   char *mot_clau;
   char *mot_actual;
   char mot_tmp[BUFSIZ];
   char ** noms_semblants;
   int *distancies_semblants;
   int num_paraules;
   int digits_ok;
   int max_distancia;
   int pos_max_distancia;
   int num_actual;
   int resultat[1];
   int i;
   
   if (argc != 4) {
      imprimir_sintaxis(stdout,1);
      return 1;
   }
   fitxer = fopen(argv[1],"r");
   if (fitxer == NULL) {
      imprimir_sintaxis(stdout,1);
      return 1;
   }
   mot_clau = (char *)malloc(TAM_MAX_PARAULA * sizeof(char));
   strcpy(mot_clau,argv[2]);
   for (i = strlen(argv[2]); i < TAM_MAX_PARAULA; i++) {
      mot_clau[i] = ' ';
   }
   digits_ok = 1;
   for (i = 0; i < strlen(argv[3]); i++) {
      if (!isdigit(argv[3][i])) {
         digits_ok = 0;
         break;
      }
   }
   if (!digits_ok) {
      imprimir_sintaxis(stdout,1);
      return -1;
   }
   num_paraules = atoi(argv[3]);
   noms_semblants = (char **)malloc(num_paraules * sizeof(char *));
   distancies_semblants = (int *)malloc(num_paraules * sizeof(int));
   mot_actual = (char *)malloc(TAM_MAX_PARAULA * sizeof(char));
   imprimir_sintaxis(stdout,2);
   num_actual = 0;
   max_distancia = 0;
   
   // Llegim paraula, tamany fix: TAM_MAX_PARAULA
   fscanf(fitxer,"%s", mot_tmp);
   if (strlen(mot_tmp) < TAM_MAX_PARAULA) {
      for (i = 0; i<strlen(mot_tmp); i++) {
         mot_actual[i] = mot_tmp[i];
      }
      for (i = strlen(mot_tmp); i < TAM_MAX_PARAULA; i++) {
         mot_actual[i] = ' ';
      }
   } else {
      for (i = 0; i<TAM_MAX_PARAULA; i++) {
         mot_actual[i] = mot_tmp[i];
      }
   }
   
   // Execucio de la distancia de Levenshtein per a cada paraula del fitxer i pel mot clau,
   // classificant les 'num_paraules' amb menor distancia, posteriorment guardades a fitxer.
   
   while (!feof(fitxer)) {
      CsOn();
      distancia_levenshtein(mot_clau,mot_actual,&resultat[0]);
      CsOff();
      distancia = resultat[0];
            
      if (num_actual < num_paraules) {
         if (!paraula_repetida(mot_actual,&noms_semblants[0],num_actual)) {
            noms_semblants[num_actual] = (char *)malloc(TAM_MAX_PARAULA * sizeof(char));
            strcpy(noms_semblants[num_actual],mot_actual);
            distancies_semblants[num_actual] = distancia;
            if (distancia > max_distancia) {
               max_distancia = distancia;
               pos_max_distancia = num_actual;
            }
            num_actual++;
         }
      }
      else {
         if (distancia < max_distancia) {
            if (!paraula_repetida(mot_actual,&noms_semblants[0],num_paraules)) {
               strcpy(noms_semblants[pos_max_distancia],mot_actual);
               distancies_semblants[pos_max_distancia] = distancia;
               pos_max_distancia = distancia_maxima_actual(&distancies_semblants[0],num_paraules);
               max_distancia = distancies_semblants[pos_max_distancia];
            }
         }
      }
      // Llegim paraula, tamany fix: TAM_MAX_PARAULA
      fscanf(fitxer,"%s", mot_tmp);
      if (strlen(mot_tmp) < TAM_MAX_PARAULA) {
         for (i = 0; i<strlen(mot_tmp); i++) {
            mot_actual[i] = mot_tmp[i];
         }
         for (i = strlen(mot_tmp); i < TAM_MAX_PARAULA; i++) {
            mot_actual[i] = ' ';
         }
      } else {
         for (i = 0; i<TAM_MAX_PARAULA; i++) {
            mot_actual[i] = mot_tmp[i];
         }
      }
   }
   fclose(fitxer);
   ordenar_taula_paraules(&noms_semblants[0],&distancies_semblants[0],num_paraules,num_actual);
   escriure_resultats("mots_semblants.txt",&noms_semblants[0],&distancies_semblants[0],num_paraules,num_actual,&mot_clau[0]);
   imprimir_sintaxis(stdout,3);
   free(noms_semblants);
   free(distancies_semblants);
   free(mot_clau);
   free(mot_actual);
   return 0;
}


